
#include <utils/p2c.h>
#include <graphics/newkbd.h>
#include "logglobals.h"
#include "utils.h"
#include "node.h"
#include "log.h"
#include "tool.h"

void newnptr(log_nrec **node)
{
	*node = (log_nrec *)Malloc(sizeof(log_nrec));
}

/// Create a new node.
void newnode(log_nrec **node, uchar st)
{
	newnptr(node);
	(*node)->simtype = findsimtype(st);
	(*node)->changed = false;
	(*node)->conflict = false;
	(*node)->conflict1 = false;
	(*node)->conflict2 = false;
	(*node)->oconflict = false;
	(*node)->flag = false;
	(*node)->keep = false;
	(*node)->ref = 0;
	newattrs(&(*node)->attr, (*node)->simtype->nnumattrs, (*node)->simtype->nattr);
	(*node)->info = (na_long)0;
	(*node)->temp = (na_long)0;
	stamp(&(*node)->simtype->netstamp);
	calltoolnode(*node, act_newnode);
	(*node)->next = gg.nbase;
	gg.nbase = *node;
	if (gg.traceflag)
		fprintf(tracefile, "Make node %p\n", *node);
}

/// Create a new node, copied from another.
void copynode(log_nrec *old, log_nrec **node)
{
	newnptr(node);
	**node = *old;
	(*node)->ref = 0;
	copyattrs(&(*node)->attr, old->attr, (*node)->simtype->nnumattrs,
			(*node)->simtype->nattr);
	gg.actnode2 = old;
	calltoolnode(*node, act_copynode);
	(*node)->next = gg.nbase;
	gg.nbase = *node;
	if ((*node)->conflict || (*node)->oconflict)
	{
		(*node)->confnext = nodeconflictbase;
		nodeconflictbase = *node;
	}
	if (gg.traceflag)
		fprintf(tracefile, "Copy node %p from %p\n", *node, old);
}


/// Dispose of a node.
void disposenptr(log_nrec **node)
{
	free(*node);
	*node = NULL;
}


#define rtn             "DISPNODE"


void disposenode(log_nrec **node)
{
	log_nrec *n1, *n2;

	if (gg.traceflag)
		fprintf(tracefile, "Dispose node %p\n", *node);
	if (*node == NULL)
		return;
	if ((*node)->conflict || (*node)->oconflict)
	{
		n1 = nodeconflictbase;
		n2 = NULL;
		while (n1 != NULL && n1 != *node)
		{
			n2 = n1;
			n1 = n1->confnext;
		}
		if (n1 == NULL)
			report(11, rtn);
		else if (n2 == NULL)
			nodeconflictbase = (*node)->confnext;
		else
			n2->confnext = (*node)->confnext;
	}
	if (*node == gg.probenode)
		gg.probenode = NULL;
	stamp(&(*node)->simtype->netstamp);
	calltoolnode(*node, act_disposenode);
	n1 = gg.nbase;
	if (*node == n1)
	{
		gg.nbase = (*node)->next;
	}
	else
	{
		while (n1 != NULL && n1->next != *node)
			n1 = n1->next;
		if (n1 == NULL)
			report(10, rtn);
		else
			n1->next = (*node)->next;
	}
	disposeattrs(&(*node)->attr, (*node)->simtype->nnumattrs, (*node)->simtype->nattr);
	disposenptr(node);
}

#undef rtn


int nodeexists(log_nrec *node)
{
	log_nrec *n1;

	n1 = gg.nbase;
	while (n1 != NULL && n1 != node)
		n1 = n1->next;
	return n1 != NULL;
}


void newnoderef(log_nrec **node, uchar st, long ref)
{
	newnode(node, st);
	(*node)->ref = ref;
}


log_nrec *refnode(log_nrec *node)
{
	if (node == NULL)
		return node;
	node->ref++;
	if (gg.traceflag)
		fprintf(tracefile, "Reference node %p -> %d\n", node, node->ref);
	return node;
}


#define rtn             "UNREFNODE"


void unrefnode(log_nrec **node)
{
	if (*node == NULL)
		return;
	(*node)->ref--;
	if (gg.traceflag)
		fprintf(tracefile, "Unref node %p -> %d\n", *node, (*node)->ref);
	if ((*node)->ref > 0)
		return;
	if ((*node)->ref < 0)
		report(10, rtn);
	else {
		if (!(*node)->keep)
			disposenode(node);
	}
}

#undef rtn


void switchnode(log_nrec **node, log_nrec *n2)
{
	if (*node != n2)
	{
		unrefnode(node);
		*node = refnode(n2);
	}
}

/** @brief Print a detailed dissection of the major data structures.
			For debugging only.
*/
void dumpnodes()
{
	log_nrec *n, *n1;
	log_hwrec *hw;
	log_vwrec *vw;
	log_srec *s;
	log_grec *g;
	short i, j;
	timerec timevar;
	daterec datevar;
	int savebool;
	char STR3[9];
	short FORLIM;
	log_krec *WITH;

	savebool = gg.traceflag;
	gg.traceflag = true;
	tracemessage("");
	gg.traceflag = savebool;
	sysdate(&datevar);
	systime(&timevar);
	fprintf(tracefile,
			"--------------------------------------------------------  ");
	fprintf(tracefile, "%2d/%2d/%02d", datevar.month, datevar.day, datevar.year);
	fprintf(tracefile, "  %2d:%2d\n\n", timevar.hour, timevar.minute);
	n = gg.nbase;
	fprintf(tracefile, "NODES\n");
	i = 0;
	j = 0;
	TRY(try6);
	while (j == i && n != NULL)
	{
		i++;
		fprintf(tracefile, "%p  T:%s  Ref:%d",
				n, n->simtype->name, n->ref);
		FORLIM = gg.lastsignal;
		for (j = 0; j < FORLIM; j++)
		{
			if (gg.signaltab[j].np == n)
				fprintf(tracefile, "   %s", gg.signaltab[j].name);
		}
		putc('\n', tracefile);
		TRY(try7);
		gg.actfile = gg.tracefile;
		gg.actflag = true;
		calltoolnode(n, act_writenode);
		RECOVER(try7);
		if (P_escapecode == -20)
			goto _Ltry6;
		fprintf(tracefile, "  (Failed to output node value)   %d\n",
				P_escapecode);
		ENDTRY(try7);
		n1 = gg.nbase;
		j = 1;
		while (j < i && n1 != n)
		{
			j++;
			n1 = n1->next;
		}
		if (j < i)
			fprintf(tracefile, "Infinite loop in node list\n");
		n = n->next;
	}
	RECOVER2(try6,_Ltry6);
	if (P_escapecode == -20)
		_Escape(P_escapecode);
	fprintf(tracefile, "  Internal error, %d\n", P_escapecode);
	ENDTRY(try6);
	putc('\n', tracefile);
	hw = gg.pages[gg.curpage - 1]->hwbase;
	fprintf(tracefile, "HWIRES\n");
	TRY(try8);
	while (hw != NULL)
	{
		fprintf(tracefile, "%p  X1:%4d X2:%4d  Y:%4d  N:%p  C:%s\n",
				hw, hw->x1, hw->x2, hw->y, hw->node,
				colorname(STR3, hw->wcolr));
		hw = hw->next;
	}
	RECOVER(try8);
	if (P_escapecode == -20)
		_Escape(P_escapecode);
	fprintf(tracefile, "  Internal error, %d\n", P_escapecode);
	ENDTRY(try8);
	putc('\n', tracefile);
	vw = gg.pages[gg.curpage - 1]->vwbase;
	fprintf(tracefile, "VWIRES\n");
	TRY(try9);
	while (vw != NULL)
	{
		fprintf(tracefile, "%p   X:%4d Y1:%4d Y2:%4d  N:%p  C:%s\n",
				vw, vw->x, vw->y1, vw->y2, vw->node,
				colorname(STR3, vw->wcolr));
		vw = vw->next;
	}
	RECOVER(try9);
	if (P_escapecode == -20)
		_Escape(P_escapecode);
	fprintf(tracefile, "  Internal error, %d\n", P_escapecode);
	ENDTRY(try9);
	putc('\n', tracefile);
	s = gg.pages[gg.curpage - 1]->sbase;
	fprintf(tracefile, "SOLDER\n");
	TRY(try10);
	while (s != NULL)
	{
		fprintf(tracefile, "%p   X:%4d  Y:%4d    HW: %p,%p",
				s, s->x, s->y, s->hwire, s->hwire2);
		if (s->hwire == NULL)
			fprintf(tracefile, "      ");
		fprintf(tracefile, "    VW: %p,%p\n", s->vwire, s->vwire2);
		s = s->next;
	}
	RECOVER(try10);
	if (P_escapecode == -20)
		_Escape(P_escapecode);
	fprintf(tracefile, "  Internal error, %d\n", P_escapecode);
	ENDTRY(try10);
	putc('\n', tracefile);
	g = gg.pages[gg.curpage - 1]->gbase;
	fprintf(tracefile, "GATES\n");
	TRY(try11);
	while (g != NULL)
	{
		fprintf(tracefile, "%p  G:%3d    X:%4d  Y:%4d  T:%s        %s\n",
				g, g->g, g->x, g->y, g->kind->simtype->name,
				g->kind->name);
		WITH = g->kind;
		FORLIM = WITH->numpins;
		for (i = 0; i < FORLIM; i++)
		{
			fprintf(tracefile, "              Pin %2d   X:%4d  Y:%4d  N:%p",
					i + 1, g->pinpos[i].x, g->pinpos[i].y, g->pin[i]);
			TRY(try12);
			fprintf(tracefile, " [%d]", g->pin[i]->ref);
			RECOVER(try12);
			if (P_escapecode == -20)
				goto _Ltry11;
			ENDTRY(try12);
			putc('\n', tracefile);
		}
		TRY(try13);
		gg.actfile = gg.tracefile;
		gg.actflag = true;
		calltoolgate(g, act_writegate);
		RECOVER(try13);
		if (P_escapecode == -20)
			goto _Ltry11;
		fprintf(tracefile, "  (Failed to output gate value)   %d\n",
				P_escapecode);
		ENDTRY(try13);
		g = g->next;
	}
	RECOVER2(try11,_Ltry11);
	if (P_escapecode == -20)
		_Escape(P_escapecode);
	fprintf(tracefile, "  Internal error, %d\n", P_escapecode);
	ENDTRY(try11);
	fprintf(tracefile,
			"\n\n-------------------------------------------------------------------------\n\n\n");
}

void delgetnode(log_nrec **save, log_nrec **old, log_nrec **node)
{
	copynode(*old, node);
}

#define rtn             "COMBINENODES"


/// Combine node N1 into N2. Adjust all references accordingly.
void combinenodes(log_nrec **n2, log_nrec **n1, cnrec *cnbase)
{
	log_nrec *nn1, *nn2, *n3;
	log_hwrec *hw;
	log_vwrec *vw;
	log_grec *g;
	cnrec *cn;
	log_hnrec *hn;
	log_tool *oldsimtype, *newsimtype;
	short i, pg, FORLIM, FORLIM1;

	if (gg.traceflag)
		fprintf(tracefile, "Combine node %ld into %ld\n",
				(long)(*n1), (long)(*n2));
	if (*n2 == NULL)
		switchnode(n2, *n1);
	nn1 = *n1;   /*VAR params may change out from under us*/
	nn2 = *n2;
	if (nn1 != NULL)
	{
		if (nn2 == nn1)
		{
			stamp(&nn2->simtype->netstamp);
		}
		else
		{
			if (nn1->simtype != nn2->simtype)
			{
				if (nn1->simtype->simtype >= 8)
				{
					if (nn2->simtype->simtype >= 8)
					{
						/* shouldn't happen */
						report(10, rtn);
					}
					else
					{
						n3 = nn1;
						nn1 = nn2;
						nn2 = n3;
					}
				}
				gg.actx = nn1->simtype->simtype;
				gg.actnode2 = nn1;
				calltoolnode(nn2, act_combineintonode);
			}
			else
			{
				gg.actnode2 = nn1;
				calltoolnode(nn2, act_combinenodes);
			}
			oldsimtype = nn1->simtype;
			newsimtype = nn2->simtype;
			stamp(&newsimtype->netstamp);
			nn2->changed = true;
			FORLIM = gg.numpages;
			for (pg = 0; pg < FORLIM; pg++)
			{
				hw = gg.pages[pg]->hwbase;
				while (hw != NULL)
				{
					if (hw->node == nn1)
						switchnode(&hw->node, nn2);
					hw = hw->next;
				}
				vw = gg.pages[pg]->vwbase;
				while (vw != NULL)
				{
					if (vw->node == nn1)
						switchnode(&vw->node, nn2);
					vw = vw->next;
				}
				g = gg.pages[pg]->gbase;
				while (g != NULL)
				{
					FORLIM1 = g->kind->numpins;
					for (i = 0; i < FORLIM1; i++)
					{
						if (g->pin[i] == nn1)
							switchnode(&g->pin[i], nn2);
					}
					g = g->next;
				}
			}
			FORLIM = gg.lastsignal;
			for (i = 1; i <= FORLIM; i++)
			{
				if (gg.signaltab[i - 1].np == nn1)
				{
					switchnode(&gg.signaltab[i - 1].np, nn2);
					if (oldsimtype->simtype < 8 && newsimtype->simtype >= 8)
					{
						hn = gg.hnbase;
						while (hn != NULL)
						{
							if (hn->sig == i)
								newattrs(&hn->attr, newsimtype->hnumattrs, newsimtype->hattr);
							hn = hn->next;
						}
					}
				}
			}
			cn = cnbase;
			while (cn != NULL)
			{
				if (cn->n == nn1)
					cn->n = nn2;
				cn = cn->next;
			}

			if (nodeexists(nn1))
			{
				report(11, rtn);
				disposenode(&nn1);
			}
		}
	}
	*n1 = *n2;
}

#undef rtn


/* N is one of the nodes to be merged into N2, later.  If N is nil, says
   that N2 must be assigned a node, but we don't have anything to connect
   to it yet. */

void queuecombine(cnrec **cnbase, log_nrec **n, log_nrec **n2)
{
	cnrec *cn;

	if (gg.traceflag)
		fprintf(tracefile, "Queue-combine,  n = %ld,  np -> %ld\n", (long)(*n), (long)(*n2));
	if (*n == NULL)
		return;
	cn = (cnrec *)Malloc(sizeof(cnrec));
	cn->next = *cnbase;
	*cnbase = cn;
	cn->n = *n;
	cn->np = n2;
}

void checknode(log_nrec *n, struct LOC_checkcombine *LINK)
{
	log_tool *newsimtype;
	char STR1[94];

	newsimtype = n->simtype;
	if (newsimtype->simtype < 8)
		return;

	if (LINK->cursimtype->simtype < 8)
	{
		LINK->cursimtype = newsimtype;
		return;
	}

	if (LINK->cursimtype->simtype == newsimtype->simtype)
		return;

	if (LINK->passcount == 2)
	{
		sprintf(STR1, "Tried to connect %s signal to %s",
				newsimtype->shortname, LINK->cursimtype->shortname);
		message(STR1);
	}
	LINK->good = false;
}

/* Merge all queued nodes.  If any simtype conflicts would result, don't
   merge any nodes.  If a node doesn't have anything merged into it, make
   a new node. */
int checkcombine(cnrec **cnbase_)
{
        struct LOC_checkcombine V;
        cnrec *cn;

        V.cnbase = cnbase_;
        V.passcount = 1;
        do
		{
                cn = *V.cnbase;
                while (cn != NULL)
				{
                        cn->done = false;
                        cn = cn->next;
                }
                cn = *V.cnbase;
                V.good = true;
                while (cn != NULL && V.good)
				{
                        V.cursimtype = simtype_ignore;
                        scancn(cn, &V);
                        cn = cn->next;
                }

				if (!V.good && V.passcount == 1)
				{
                        cn = *V.cnbase;
                        while (cn != NULL)
						{
                                confirmsimtype(cn->n);
                                confirmsimtype(*cn->np);
                                cn = cn->next;
                        }
                }

				if (gg.traceflag && *V.cnbase != NULL)
                        fprintf(tracefile, "Check-combine (pass %d) finds %s\n",
                                        V.passcount, V.good ? " TRUE" : "FALSE");
                V.passcount++;
        } while (!(V.good || V.passcount > 2));

		if (V.good)
		{
                cn = *V.cnbase;
                while (cn != NULL)
				{
                        combinenodes(cn->np, &cn->n, *V.cnbase);
                        cn = cn->next;
                }
        }

		while (*V.cnbase != NULL)
		{
                cn = (*V.cnbase)->next;
                Free(*V.cnbase);
                *V.cnbase = cn;
        }
        return V.good;
}


int trycombinenodes(log_nrec **n1, log_nrec **n2)
{
        cnrec *cnbase;

        cnbase = NULL;
        queuecombine(&cnbase, n1, n2);
        return (checkcombine(&cnbase));
}

void reportnodeconflict(log_nrec *n)
{
	if (n->conflict || !conflictenbl)
		return;
	if (!n->oconflict)
	{
		n->confnext = nodeconflictbase;
		nodeconflictbase = n;
		n->conflict1 = (3L - conflictdelay) & 1;
		n->conflict2 = ((3L - conflictdelay) / 2) & 1;
	}
	n->conflict = true;
}

void getnodeval(log_nrec *n, double *val, char *opts)
{
	gg.actval = 0.0;
	strcpy(gg.actstr, opts);
	calltoolnode(n, act_nodeval);
	*val = gg.actval;
}

