#include <utils/p2c.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <utils/strings.h>
#include "log.h"
#include "logglobals.h"
#include "utils.h"

void getword(char *buf, char *wrd)
{
	size_t len = strlen(buf);
	char *str1 = (char *) malloc(sizeof(char) * (len + 1));
	char *str2 = (char *) malloc(sizeof(char) * (len + 1));
	strcpy(str1, strltrim(strrtrim(strcpy(str2, buf))));
	strcpy(buf, str1);
	free(str1);
	free(str2);

	if (*buf == '"')
		strwordx(buf, wrd);
	else
	{
		strwordx(buf, wrd);
		strupper(wrd, wrd);
	}
}

void stamp(long *i)
{
	gg.curstamp++;
	*i = gg.curstamp;
}

void dounits(char *s, double *r)
{
	char ch;

	if (*s == '\0')
		return;
	ch = s[0];
	if (ch >= 'a')
		ch -= 32;
	if (ch == 'F')
	{
		*r *= 1e-15;
		return;
	}
	if (ch == 'P')
	{
		*r *= 1e-12;
		return;
	}
	if (ch == 'N')
	{
		*r *= 1e-9;
		return;
	}
	if (ch == 'U')
	{
		*r *= 1e-6;
		return;
	}
	if (ch == 'K')
	{
		*r *= 1e3;
		return;
	}
	if (ch == 'G')
	{
		*r *= 1e9;
		return;
	}
	if (ch == 'T')
	{
		*r *= 1e12;
		return;
	}
	if (ch != 'M')
		return;
	if (strlen(s) >= 3 && (s[1] == 'E' || s[1] == 'e') &&
			(s[2] == 'G' || s[2] == 'g'))
		*r *= 1e6;
	else
		*r *= 1e-3;
}


void readreal(char *s_, double *r)
{
	char s[256];
	long i;
	char *str1;
	char str2[256];

	strcpy(s, s_);
	if (!(*s != '\0' && (s[0] == '.' || s[0] == '+' || s[0] == '-' || isdigit(s[0]))))
		return;
	*r = strtod(s, &str1);
	i = str1 - s + 1;
	strcpy_overlap(s, s + i - 1);
	strcpy(str2, strltrim(s));
	strcpy(s, str2);
	dounits(s, r);
}

/// Put two shortint variables into increasing numerical order.
void sortshints(short *x, short *y)
{
	short temp;

	if (*x <= *y)
		return;
	temp = *x;
	*x = *y;
	*y = temp;
}

/// Convert a real number to a string.
char *realstr(char *Result, double r, short p)
{
	size_t i, j;
	char s[81];

	s[80] = '\0';
	if ((fabs(r) < 1e-2 || fabs(r) >= 1e6) && r != 0)
	{
		sprintf(s, "% .13E", r);
		i = strlen(s) + 1;
		s[i - 1] = '\0';
		i = strposc(s, 'E', 1L) - 1;
		j = strposc(s, '.', 1L);
		while (s[i - 1] == '0' || (p != 0 && i > j + p))
		{
			strcpy_overlap(s + i - 1, s + i);
			i--;
		}
		if (s[i - 1] == '.')
		{
			strcpy_overlap(s + i - 1, s + i);
			i--;
		}
		while (strlen(s) > i + 3 && s[i + 2] == '0')
			strcpy_overlap(s + i + 2, s + i + 3);
		if (s[i + 1] == '+')
			strcpy_overlap(s + i + 1, s + i + 2);
		return strcpy(Result, strltrim(s));
	}

	if (p == 0)
	{
		sprintf(s, "%30.9f", r);
		i = strlen(s) + 1;
		do
		{
			i--;
		} while (s[i - 1] == '0');

		if (s[i - 1] == '.')
			i--;
		s[i] = '\0';
		return strcpy(Result, strltrim(s));
	}
	sprintf(s, "%25.*f", p, r);
	i = strlen(s) + 1;
	s[i - 1] = '\0';
	/* zfprintf(stdout, "realstr:%lf p: %d s:  %s\n", r, p, s); **MDG** */
	return strcpy(Result, strltrim(s));
}


/// Convert a unitted real number to a string.
char *realunit(char *Result, double r, short p, char *u, int mu)
{
	char s[81], STR1[81];
	char STR2[162];

	/* zfprintf(stdout, "Realunit r: %le\n ", r);  ***MDG** */
	*s = '\0';
	if (r == 0 || fabs(r) >= 1.0e15 || fabs(r) < 1.0e-16)
	{
		*s = '\0';
	}
	else if (fabs(r) >= 1e12)
	{
		strcpy(s, "T");
		r /= 1e12;
	}
	else if (fabs(r) >= 1e9)
	{
		strcpy(s, "G");
		r /= 1e9;
	}
	else if (fabs(r) >= 1e6)
	{
		strcpy(s, "Meg");
		r /= 1e6;
	}
	else if (fabs(r) >= 1e3)
	{
		strcpy(s, "K");
		r /= 1e3;
	}
	else if (fabs(r) >= 0.1)
	{
		*s = '\0';
	}
	else if (fabs(r) >= 1e-3)
	{
		strcpy(s, "m");
		r *= 1e3;
	}
	else if (fabs(r) >= 1e-6)
	{
		strcpy(s, "u");
		r *= 1e6;
	}
	else if (fabs(r) >= 1e-9)
	{
		strcpy(s, "n");
		r *= 1e9;
	}
	else if (fabs(r) >= 1e-12)
	{
		strcpy(s, "p");
		r *= 1e12;
	}
	else
	{
		strcpy(s, "f");
		r *= 1e15;
	}

	sprintf(s, "%s%s", realstr(STR1, r, p), strcpy(STR2, s));
	sprintf(Result, "%s%s", s, u);
	/* zfprintf(stdout, "realunit Result: %s\n", Result); ***MDG** */
	return Result;
}

long getint(char *s, long def)
{
	char w[10];
	size_t i;
	long j, sign;
	int valid;
	size_t FORLIM;

	getword(s, w);
	if (*w == '-')
	{
		strcpy_overlap(w, w + 1);
		sign = -1;
	}
	else
	{
		sign = 1;
	}
	j = 0;
	valid = true;
	FORLIM = strlen(w);
	for (i = 0; i < FORLIM; i++)
	{
		if (w[i] < '0' || w[i] > '9')
			valid = false;
		else
			j = j * 10 + w[i] - 48;
	}

	if (*w != '\0' && valid)
		return (j * sign);
	else
		return def;
}


void pgetint(char *buf, long *i, long def)
{
	*i = getint(buf, def);
}


double getreal(char *s, double def)
{
	char w[81];
	double r;

	getword(s, w);
	r = def;
	TRY(try1);
	readreal(w, &r);
	RECOVER(try1);
	if (P_escapecode == -20)
		_Escape(P_escapecode);
	r = def;
	ENDTRY(try1);
	return r;
}

void pgetreal(char *buf, double *r, double def)
{
	*r = getreal(buf, def);
}


void getbool(char *buf, int *b)
{
	char w[17];

	getword(buf, w);
	if (!strcmp(w, "ON") || !strcmp(w, "YES") || !strcmp(w, "Y") ||
			!strcmp(w, "RABBITS"))
	{   /*avoid*/
		*b = true;
		return;
	}
	if (!strcmp(w, "OFF") || !strcmp(w, "NO") || !strcmp(w, "N"))
	{
		*b = false;
		return;
	}

	if (!strcmp(w, "?"))
		tempverbose = true;
	else if (*w == '\0')
		*b = !*b;
}

